import os

from flask import render_template, g

from cbr_website_beta.apps.CBR_Config                   import cbr_config
from cbr_website_beta.apps.llms                         import blueprint
from cbr_website_beta.apps.llms.Prompt_Examples         import Prompt_Examples
from cbr_website_beta.aws.s3.DB_Users                   import DB_Users
from cbr_website_beta.flask.decorators.allow_annonymous import admin_only
from cbr_website_beta.utils.Web_Utils                   import Web_Utils
from osbot_utils.utils.Env import in_github_action

EXPECTED_ROUTES__LLMS = ['/llms/dev/just-chat-ui'     ,
                         '/llms/dev/multiple-athenas' ,
                         '/llms/dev/no-system-prompt' ]
INTRO_TO_USER_DATA     = "We have asked the user to provide some information and included below is the data provided, please customise the answers as much as possible to these user preferences:\n"


@blueprint.route('/dev/no-system-prompt')
@admin_only
def no_system_prompt():
    kwargs = dict(examples_title        = 'Prompt examples'                                          ,
                  examples_texts        = Prompt_Examples().no_system_prompt()                       ,  # todo: improve caching of Prompt_Examples
                  title                 = "Dev | FMs with no system prompt"                          ,
                  url_athena            = target_athena_url() + '/open_ai/prompt_with_system__stream',  # todo: refactor into helper method
                  content_view          = '/llms/dev/no_prompt.html'                                 ,
                  template_name_or_list = '/pages/page_with_view.html'                               )
    return render_template(**kwargs)

@blueprint.route('/dev/just-chat-ui')
@admin_only
def just_chat_ui():
    kwargs = dict(examples_title='Just Chat UI',
                  examples_texts=Prompt_Examples().no_system_prompt(),  # todo: improve caching of Prompt_Examples
                  title="Dev | FMs with no system prompt",
                  url_athena=target_athena_url(),
                  content_view='/llms/dev/no_prompt.html',
                  template_name_or_list='/layouts/empty-page.html')
    return render_template(**kwargs)

@blueprint.route('/dev/multiple-athenas')
@admin_only
def multiple_athenas():
    kwargs = dict(examples_title='Prompt examples',
                  examples_texts=Prompt_Examples().no_system_prompt(),  # todo: improve caching of Prompt_Examples
                  title="Athena (model 4.5)",
                  url_athena=target_athena_url() + '/open_ai/prompt_with_system__stream',
                  content_view='/llms/dev//multiple-athenas.html',
                  #template_name_or_list='/pages/page_with_view.html',
                  template_name_or_list='/layouts/just-top-banner.html'
                  )
    return render_template(**kwargs)
    #return render_template('/llms/open_ai/osbot_llm.html', user_data=user_data_for_prompt())



# previous version of the routes

# @blueprint.route('/chat-gpt/ui-simple')
# def ui_simple():
#     questions = ['hi', 'who are you?', 'what questions should I ask my CISO?', 'do you have siblings?']
#     return render_template('llms/chat_gpt/ui_simple.html', questions=questions)
#
# @blueprint.route('/chat-gpt/ui-with-prompts')
# def ui_with_prompts():
#     athena_rest_api = Athena_Rest_API()
#     athena_prompt   = athena_rest_api.athena_prompt()
#
#     questions     = ['hi', 'who are you?', 'what questions should I ask my CISO?', 'do you have siblings?']
#     user_data     = user_data_for_prompt()
#     system_prompt = athena_prompt
#     return render_template('llms/chat_gpt/ui_with_prompts.html', questions=questions, user_data=user_data, system_prompt=system_prompt)
#
# @blueprint.route('/chat-gpt/prompt', methods=['POST'])
# def prompt():
#     open_api     = API_Open_AI()
#     request_form = request.form.to_dict()
#     prompt       = request_form.get('prompt')
#     answer       = open_api.ask_athena_one_question(prompt)
#     ui_data      = {'response':answer}
#     return jsonify(ui_data)
#
# @blueprint.route('/chat-gpt/question-with-prompt', methods=['POST'])
# def question_with_prompt():
#     request_form  = request.form.to_dict()
#     user_data     = request_form.get('user_data'    )
#     user_question = request_form.get('user_question')
#     system_prompt = request_form.get('system_prompt')
#     user_history  = json_parse(request_form.get('user_history' , "[]"))
#     #user_history = []
#
#     #full_prompt = f"---SYSTEM--PROMPT---\n\n {system_prompt}\n\n ---USER--QUESTION---\n\n {user_question}\n\n ---USER--DATA---\n\n {user_data}\n\n"
#     print('user_history', user_history)
#     open_api = API_Open_AI()
#     answer = open_api.ask_question_with_user_data_and_prompt(user_question, user_data, system_prompt, user_history)
#     #system_prompt = replace(system_prompt, '\n', '<br/>')
#     answer = replace(answer, '\n', '<br/>')
#     ui_data      = {'response': answer}
#     return jsonify(ui_data)


# def user_data_for_prompt():
#     raw_user_data = current_user_data()
#     vars_to_add = ['user_id', 'name', 'title', 'email', 'organisation', 'sector', 'size of organisation', 'country', 'linkedin', 'Profile', 'Prompt Helper']
#     user_data = f"This is info about the user from its profile:\n"
#     for var in vars_to_add:
#         value = raw_user_data.get(var)
#         if value:
#             user_data += f"\n{var}: {value}"
#     return user_data




## helper methods : todo: refactor into separate class
def target_athena_url():
    env           = cbr_config.env()
    target_server = "https://athena.thecyberboardroom.com"
    dev_server    = 'https://athena-dev.cyber-boardroom.com'
    local_server  = 'http://localhost:5111'

    if in_github_action():
        target_server = 'https://athena-dev.cyber-boardroom.com'
    else:
        if env   == 'LOCAL': target_server = dev_server  # local_server  # todo change the way the target server is mapped and consumed in cbr website
        elif env == 'DEV'  : target_server = dev_server

    return f"{target_server}"

def user_data_for_prompt():
    raw_user_data = current_user_data()
    if raw_user_data == {} or raw_user_data.get('user_id') is None:
        return ""
    vars_to_add = [
        'First name', 'Last name', 'Role', 'Organisation',
        'Sector', 'Size of organisation', 'Country', 'Linkedin',
        'Additional suggested prompts for Athena, your AI advisor' ]

    # Determine the longest variable name for proper alignment
    longest_var = max(len(var) for var in vars_to_add)

    lines_of_data = [INTRO_TO_USER_DATA]

    # Format the data into a readable list, aligned in columns
    for var in vars_to_add:
        value = raw_user_data.get(var)
        if value:
            # Capitalize each word in var_name for better presentation
            var_name_capitalized = ' '.join(word.capitalize() for word in var.split())
            # Align the data with the longest variable name
            line = f"{var_name_capitalized:<{longest_var + 2}}: {value}"
            lines_of_data.append(line)

    # Join the lines into a single string with newline separation
    user_data = "\n".join(lines_of_data)
    return user_data


def current_user_data():
    if cbr_config.login_disabled():
        return {}
    db_users = DB_Users()
    #user_data = Current_User().user_data()
    user_data = g.user_data
    user_id = user_data.get('username')
    if user_id:
        db_user = db_users.db_user(user_id)
        if db_user.exists() is False:
            print(f"Creating user: {user_id}")
            db_user.create()
        return db_user.metadata()
    return {}








import logging

from flask import g
from cbr_website_beta.utils.Site_Utils import Site_Utils

LOGGING_LEVEL_FOR_G     = logging.DEBUG
#LOGGING_FORMAT_FOR_G    = '%(levelname)s: %(message)s'
LOGGING_FORMAT_FOR_G    = '%(message)s'
LOGGING_FORMATTER_FOR_G = logging.Formatter(LOGGING_FORMAT_FOR_G)

from flask.globals import _app_ctx_stack

class Logging_in_G(logging.Handler):
    def emit(self, record):
        if _app_ctx_stack.top is not None:
            if hasattr(g, 'log_list'):
                g.log_list = getattr(g, 'log_list', [])
                #record['level'] = self.level
                g.log_req_count = len(g.log_list_all)

                if type(record.msg) is dict:
                    record.msg['level'] = record.levelname
                    record.msg['index'] = g.log_req_count
                g.log_list.append(record.msg)
                g.log_list_all.append(record.msg)



def register_logging_in_g(app):
    if Site_Utils.running_in_pytest():
        return

    logging_in_handler = Logging_in_G()
    logging_in_handler.setLevel     (LOGGING_LEVEL_FOR_G   )
    logging_in_handler.setFormatter(LOGGING_FORMATTER_FOR_G)
    app.logger.addHandler(logging_in_handler)
    return app


from cbr_website_beta.apps.home.home_routes import dydb_cbr_logging
from cbr_website_beta.bots.Athena_Rest_API import Athena_Rest_API
from cbr_website_beta.utils.Markdown_Parser import Markdown_Parser
from osbot_utils.utils.Dev import pprint


class Athena_Html_Content:

    filter_name = 'athena_html_content'

    def __init__(self, app):
        app.jinja_env.filters[self.filter_name] = self.athena_html_content # todo: find a better way to register these filters

    def athena_html_content(self, path):
        try:
            athena_rest_api = Athena_Rest_API()
            content         = athena_rest_api.requests_get(path)
            return Markdown_Parser().content_to_html(content)
        except Exception as e:
            dydb_cbr_logging.log_exception(e)
            return "Error in loading athena content"



import socket
import threading
from unittest       import TestCase
from urllib.parse import urljoin

import requests
from flask          import Flask
from flask_socketio import SocketIO

from cbr_website_beta.aws.apigateway.web_sockets.server.WS__Flask_Server import WS__Flask_Server
from osbot_utils.utils.Dev import pprint
from osbot_utils.utils.Http import wait_for_port, wait_for_http
from osbot_utils.utils.Objects import obj_info


class TestCase__WS_Server(TestCase):
    ws_flask      : WS__Flask_Server
    app           : Flask
    socketio      : SocketIO
    server_thread : threading.Thread
    port          : int

    @classmethod
    def setUpClass(cls):
        cls.ws_flask      = WS__Flask_Server().setup()
        cls.app           = cls.ws_flask.app()
        cls.socketio      = cls.ws_flask.socketio()
        cls.port          = cls.find_free_port()
        cls.server_thread = threading.Thread(target=cls.socketio.run, args=(cls.app,), kwargs={'port': cls.port})
        cls.server_thread.daemon = True  # This allows the thread to be killed when the main thread exits
        cls.server_thread.start()
        wait_for_port('127.0.0.1',cls.port)

    @classmethod
    def tearDownClass(cls):
        pass
        #cls.socketio.stop()
        #cls.server_thread.join()

    @staticmethod
    def find_free_port():
        with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
            s.bind(('', 0))
            return s.getsockname()[1]

    def http_get(self, path):
        url = urljoin(self.url(), path)
        return requests.get(url).text

    def url(self):
        return f'http://localhost:{self.port}'


import os

from osbot_utils.utils.Misc import list_set
from cbr_website_beta.utils._for_OSBot_Utils import convert_paths_into_folder_dict
from osbot_utils.utils.Files import path_combine, files_list


class Flask_App_Urls:

    def __init__(self, app):
        self.app = app

    # get the list of all possible virtual paths to files that exists on disk
    def paths_to_static_pages(self, start_from=None):
        template_directory = self.template_folder()
        all_files          = files_list(template_directory)
        all_possible_paths = set()

        for file in all_files:
            # Remove the prefix of the template directory and the '.html' suffix
            path = os.path.relpath(file, template_directory)[:-5].replace("\\", "/")  # Replace backslashes on Windows
            original_path = path

            # Rule 1
            if path == 'athena':
                all_possible_paths.add('athena')
                path = 'athena/index'
            # Rule 2
            elif path == 'content':
                all_possible_paths.add('content')
                path = 'content/index'
            # Rule 3
            if path.endswith('/'):
                path += 'index'

            # Rule 3 (handle bug) # todo: fix this bug
            if start_from =='home' and path=='includes/home':           # handle coner case cause by the fact that the template name is the same as the folder (we need a better way to map these files and the attack surface)
                all_possible_paths.add('/home')
                all_possible_paths.add('/home.html')
                continue
            # Add both the URL form and the .html form for each path
            if start_from:
                if start_from not in path:
                    continue
                else:
                    path          = path         [len(start_from):]
                    original_path = original_path[len(start_from):]
            all_possible_paths.add(path)
            all_possible_paths.add(original_path + '.html')

        return list_set(all_possible_paths)

    def paths_to_static_pages_as_dict(self):
        return convert_paths_into_folder_dict(self.paths_to_static_pages())

    def root_path(self):
        return self.app.root_path

    def template_folder(self):
        return path_combine(self.root_path(), self.app.template_folder)


from flask import request, g

from cbr_website_beta.flask.Flask_Utils import Flask_Utils

MAX_COOKIE_SIZE = 25

#@xray_trace("add_cookies_to_g")
def add_cookies_to_g():
    if Flask_Utils.is_static_request():
        return
    cookie_dict = {}
    session_id = None
    for key, value in request.cookies.items():
        if len(value) > MAX_COOKIE_SIZE:
            formatted_value = f"{value[:MAX_COOKIE_SIZE]} - size: {len(value)}"
        else:
            formatted_value = value

        cookie_dict[key] = formatted_value
        if key == 'CBR_TOKEN':
            session_id = value

    g.cookies    = cookie_dict
    g.session_id = session_id


from cbr_website_beta.aws.apigateway.web_sockets.CBR__WS__DyDB  import CBR__WS__DyDB
from osbot_aws.aws.apigateway.API_Gateway_Management_API        import API_Gateway_Management_API
from osbot_utils.decorators.methods.cache_on_self               import cache_on_self
from osbot_aws.AWS_Config                                       import AWS_Config
from osbot_utils.base_classes.Kwargs_To_Self                    import Kwargs_To_Self
from osbot_utils.utils.Lists                                    import list_group_by
from osbot_utils.utils.Misc                                     import timestamp_to_str_time, timestamp_utc_now

class WS__Users(Kwargs_To_Self):
    dydb_ws     : CBR__WS__DyDB
    aws_config  : AWS_Config
    api_gateway : API_Gateway_Management_API

    def __init__(self):
        super().__init__()
        self.api_gateway = API_Gateway_Management_API(endpoint_url=self.boto3_endpoint_url())

    def active_connections__docs_metadata(self):
        try:
            return  self.dydb_ws.query_index('status','S', 'active')

        except Exception as e:
            print(f'Exception: {e}')
            return []

    def active_connections(self):
        docs_metadata = self.active_connections__docs_metadata()
        return [document.get('id') for document in docs_metadata]

    def action_connections__by_name(self):
        return list_group_by(values=self.active_connections__docs_metadata(), group_by='connection_name')

    def connection_id__for_name(self, name):
        name_connection_ids = self.connection_ids__for_name(name)
        if len(name_connection_ids) > 0:                    # if there are connection_ids with this name
            return name_connection_ids[0]                   # return  the first one

    def connection_ids__for_name(self, name):
        user_connections = self.action_connections__by_name().get(name, [])
        return [connection.get('id') for connection in user_connections]

    def user_messages(self, connection_id):
        return self.dydb_ws.document(connection_id)

    def send_to_active_connections(self, topic, data):
        active_connections = self.active_connections()
        print(active_connections)
        for connection_id in active_connections:
            self.send_ws_message(connection_id, topic, data)

            #with Duration(prefix=f'Sending messager to  on {connection_id}'):
            #dydb_document = self.dydb_ws.dydb_document(document_id=connection_id,load_data=False)
            #dydb_document.add_to_list('messages', {'sent_to_active_connections': data})


        return {'status': 'ok', 'data': data, 'active_connections': active_connections}

    def send_ws_message(self, connection_id, topic, data):
        try:
            timestamp = timestamp_utc_now()
            client_message = dict(when  = timestamp_to_str_time(timestamp),
                                  topic = topic                           ,
                                  data  = data                            )
            self.api_gateway.post_to_connection(connection_id, client_message)
            print(f'sent ws message to: {connection_id}')
            return True
        except Exception as e:
            print(f'[ERROR] failed to send  ws message to: {connection_id} | {e}')
            return False

    def send_ws_message__to_name(self,target_name, topic, data):
        connection_ids = self.connection_ids__for_name(target_name)
        if connection_ids:
            for connection_id in connection_ids:
                self.send_ws_message(connection_id, topic, data)
                print(f"message sent to '{target_name}'")
            return
        print(f"no connection found with name '{target_name}'")

    def api_gw__id(self):
        return 'ow2twc6ee4'  # todo : move to env variable

    def api_gw__region_name(self):
        return self.aws_config.region_name()

    def api_gw__stage(self):
        return 'prod'

    @cache_on_self
    def boto3_endpoint_url(self):

        endpoint_url = f'https://{self.api_gw__id()}.execute-api.{self.api_gw__region_name()}.amazonaws.com/{self.api_gw__stage()}'
        return endpoint_url

    def wss_endpoint_url(self):
        endpoint_url = f"wss://{self.api_gw__id()}.execute-api.{self.api_gw__region_name()}.amazonaws.com/{self.api_gw__stage()}"
        return endpoint_url



import frontmatter
from markdown import Markdown


class Markdown_Parser:
    def __init__(self):
        self.md = Markdown()

    def markdown_to_html(self, markdown_text):
        return self.md.convert(markdown_text)

    def content_to_html(self, content):
        metadata, markdown = frontmatter.parse(content)
        return  self.markdown_to_html(markdown)